#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# License: GNU General Public License v2
#
# Author: thl-cmk[at]outlook[dot]com
# URL   : https://thl-cmk.hopto.org
# Date  : 2021-08-29
#
# include file, will be used with (inv_)bgp_peer, (inv_)cisco_bgp_peer, (inv_)juniper_bgp_peer, huawei_bgp_peer
#
# 2022-04-17: added peer_unavail_reason/peer_unavail_reason_str for huawei bgp peers
# 2022-04-29: added device_admin_limit
# 2022-04-30: code cleanup/streamlining
# 2022-05-09: added item to BgpPeer class, this is used in the discovery function
# 2022-05-11: changed bgp_get_peer_entry to get proper parameters instead of Nontransparent list
#             added remote_as to BgpPeerItem
# 2022-05-12: merged bgp_get_ip_address_from_oid with bgp_render_ip_address
#             changed IPv6 address format to lower case as required by rfc5952 section-4.3 (this affects IPv6 items)
# 2023-01-20: add description to BgpPeerItem (for Arista)

#
# https://ftp.ripe.net/ripe/asnames/asn.txt AS Name List
#

import requests
import json
import ipaddress
from typing import List, Tuple, Dict, TypedDict
from dataclasses import dataclass


class BgpPeerItem(TypedDict):
    remote_address: str
    remote_as: str
    address_family: str
    routing_instance: str


@dataclass
class BgpPeer:
    # peer_statestr: str
    accepted_prefixes: int | None
    admin_state: int | None
    admin_statestr: str
    description: str | None
    device_admin_limit: bool | None
    fsm_established_time: int | None
    item: BgpPeerItem
    local_addr: str | None
    metric_count: List[Tuple[str, int]]
    metric_rate: List[Tuple[str, int]]
    peer_state: int
    peer_unavail_reason: int | None
    peer_unavail_reason_str: str | None
    prefix_admin_limit: int | None
    prefix_clear_threshold: int | None
    prefix_threshold: int | None
    remote_as: str | None
    remote_addr: str | None
    remote_id: str | None
    local_as: str | None
    local_id: str | None


class InvBgpPeer(TypedDict):
    remote_addr: str
    remote_id: str
    version: str
    local_addr: str
    remote_as: str
    local_as: str
    local_id: str
    bgp_type: str
    peer_state: int
    last_error_code: str
    last_error: str
    status_columns: Dict[str, str]
    address_family:  str | None
    fsm_established_time: int | None


def sec2hr(seconds):
    m, s = divmod(seconds, 60)
    h, m = divmod(m, 60)
    d, h = divmod(h, 24)
    if d > 0:
        return f'{d}d {h:02.0f}:{m:02.0f}:{s:02.0f}'
    else:
        return f'{h:02.0f}:{m:02.0f}:{s:02.0f}'


def bgp_peerstate(st):
    names = {
        0: 'none',
        1: 'idle',
        2: 'connect',
        3: 'active',
        4: 'opensnet',
        5: 'openconfirm',
        6: 'established'
    }
    return names.get(st, st)


def bgp_adminstate(st):
    names = {
        1: 'shutdown',
        2: 'enabled',
    }
    return names.get(st, st)


def get_bgp_type(local_as: str, remote_as: str) -> str:
    bgp_type = 'N/A'
    if local_as.isdigit() and remote_as.isdigit():
        if local_as == remote_as:
            bgp_type = 'iBGP'
        else:
            bgp_type = 'eBGP'

    return bgp_type


def bgp_error_code_as_hex(error_code: List[int]):
    return ''.join([f'{m:02x}' for m in error_code])


def bgp_error_as_string(error_code: List[int]):
    # byte1, byte2 = ByteToHex(bytestring).split()
    try:
        main_code, sub_code = error_code
    except ValueError:
        return 'Unknown'

    names = {
        0: {0: 'NO ERROR'},
        1: {
            0: 'Message',
            2: 'Connection Not Synchronized',
            3: 'Bad Message Length',
            4: 'Bad Message Type',
        },
        2: {
            0: 'OPEN',
            1: 'Unsupported Version Number',
            2: 'Bad Peer AS',
            3: 'Bad BGP Identifier',
            4: 'Unsupported Optional Parameter',
            5: 'Authentication Failure',
            6: 'Unacceptable Hold',
        },
        3: {
            0: 'UPDATE',
            1: 'Malformed Attribute List',
            2: 'Unrecognized Well-known Attribute',
            3: 'Missing Well-known Attribute',
            4: 'Attribute Flags Error',
            5: 'Attribute Length Error',
            6: 'Invalid ORIGIN Attribute',
            7: 'AS Routing Loop',
            8: 'Invalid NEXT_HOP Attribute',
            9: 'Optional Attribute Error',
            10: 'Invalid Network Field',
            11: 'Malformed AS_PATH',
        },
        4: {0: 'Hold Timer Expired', },
        5: {0: 'Finite State Machine Error', },
        6: {
            0: 'Administratively Shutdown',
            1: 'Max Prefix Reached',
            2: 'Peer Unconfigured',
            3: 'Administratively Reset',
            4: 'Connection Rejected',
            5: 'Other Configuration Change',
        }}

    return names[main_code].get(sub_code)


def bgp_render_ip_address(addr_type: str, addr: List[int]) -> str:
    if addr_type == '1':
        return '.'.join([str(m) for m in addr])
    elif addr_type == '2':
        # IPv6 address from snmp oid in decimal
        # [10, 1, 7, 40, 0, 0, 50, 0, 0, 0, 0, 0, 0, 0, 5, 25]
        # change to hex with leading zero
        # ['0a', '01', '07', '28', '00', '00', '32', '00', '00', '00', '00', '00', '00', '00', '05', '19']
        remote_address = ':'.join([f'{m:02x}' for m in addr]).split(':')
        # convert to long ipv6 address 0a01:0728:0000:3200:0000:0000:0000:0519
        remote_address = ':'.join([''.join(remote_address[i:i + 2]) for i in range(0, len(remote_address), 2)])
        # convert to short ipv6 address a01:728:0:3200::519
        remote_address = str(ipaddress.ip_address(remote_address))
        # replace bytes in entry with ip address
        return remote_address
    else:
        return 'N/A'


def bgp_get_ip_address_from_oid(oid_end):
    # returns peer address string from OID_END
    # u'1.4.217.119.208.34.1.1' --> 217.119.208.34
    # u'2.16.42.5.87.192.0.0.255.255.0.0.0.0.0.0.0.17.2.1' -->  42.5.87.192.0.0.255.255.0.0.0.0.0.0.0.17
    ip_address = ''
    oid_end = oid_end.split('.')
    addr_type = oid_end[0]
    addr_length = oid_end[1]

    if addr_type == '1' and addr_length == '4':  # length of ip address
        ip_address = '.'.join(oid_end[2:6])  # ipv4 address
    elif addr_type == '2' and addr_length == '16':  # ipv6 address
        ip_address = []
        # change list of str to lis of int
        ip_address += [int(m) for m in oid_end[2:18]]
        ip_address = bgp_render_ip_address(addr_type, ip_address)

    return ip_address


def bgp_get_peer_entry(
        peer_state: str,
        remote_addr: str,
        remote_as: str,
        fsm_established_transitions: str,
        admin_state: str | None = None,
        description: str | None = None,
        fsm_established_time: str | None = None,
        in_messages: str | None = None,
        in_update_elapsed_time: str | None = None,
        in_updates: str | None = None,
        local_addr: str | None = None,
        local_as: str | None = None,
        local_id: str | None = None,
        out_messages: str | None = None,
        out_updates: str | None = None,
        remote_id: str | None = None,

) -> Dict[str, BgpPeer] | None:

    bgp_peer = BgpPeer(
        peer_state=int(peer_state),
        # peer_statestr=bgp_peerstate(int(peer_state)),
        admin_state=int(admin_state) if admin_state.isdigit() else None,
        admin_statestr=bgp_adminstate(int(admin_state) if admin_state.isdigit() else None),
        fsm_established_time=int(fsm_established_time) if fsm_established_time.isdigit() else None,
        metric_count=[],
        metric_rate=[],
        item={
            'remote_address': remote_addr,
            'remote_as': f'AS{remote_as}',
            'address_family': '',
            'routing_instance': '',
        },
        device_admin_limit=None,
        prefix_admin_limit=None,
        prefix_threshold=None,
        prefix_clear_threshold=None,
        accepted_prefixes=None,
        peer_unavail_reason=0,
        peer_unavail_reason_str='',
        description=description if description else '',
        local_addr=local_addr,
        remote_as=remote_as,
        remote_addr=remote_addr,
        remote_id=remote_id,
        local_as=local_as,
        local_id=local_id,
    )

    for key, value in [
        ('inupdates', in_updates),
        ('outupdates', out_updates),
        ('intotalmessages', in_messages),
        ('outtotalmessages', out_messages),
    ]:
        try:
            bgp_peer.metric_rate.append((key, int(value)))
        except (ValueError, TypeError):
            pass

    for key, value in [
        ('fsmestablishedtransitions', fsm_established_transitions),
        ('inupdateelapsedtime', in_update_elapsed_time),
    ]:
        try:
            bgp_peer.metric_count.append((key, int(value)))
        except (ValueError, TypeError):
            pass

    return {remote_addr: bgp_peer}


def _fetch_data(url: str, timeout: int) -> dict:
    response = requests.get(
        url=url,
        timeout=timeout,
    )
    # ToDo: improve/Implement error handling
    if response.status_code == 200:
        return json.loads(response.text)
    else:
        return {}


class BgpWhois:
    __rdap_boot_strap = {
        'source': 'https://data.iana.org/rdap/asn.json',
        'description': 'RDAP bootstrap file for Autonomous System Number allocations',
        'publication': '2021-12-07T20:00:01Z',
        'services': [
            [
                [
                    '36864-37887',
                    '327680-328703',
                    '328704-329727'
                ],
                [
                    'https://rdap.afrinic.net/rdap/',
                    'http://rdap.afrinic.net/rdap/'
                ]
            ],
            [
                [
                    '4608-4865',
                    '7467-7722',
                    '9216-10239',
                    '17408-18431',
                    '23552-24575',
                    '37888-38911',
                    '45056-46079',
                    '55296-56319',
                    '58368-59391',
                    '63488-63999',
                    '64000-64098',
                    '64297-64395',
                    '131072-132095',
                    '132096-133119',
                    '133120-133631',
                    '133632-134556',
                    '134557-135580',
                    '135581-136505',
                    '136506-137529',
                    '137530-138553',
                    '138554-139577',
                    '139578-140601',
                    '140602-141625',
                    '141626-142649',
                    '142650-143673',
                    '143674-144697',
                    '144698-145721',
                    '145722-146745',
                    '146746-147769',
                    '147770-148793',
                    '148794-149817',
                    '149818-150841',
                    '150842-151865'
                ],
                [
                    'https://rdap.apnic.net/'
                ]
            ],
            [
                [
                    '1-1876',
                    '1902-2042',
                    '2044-2046',
                    '2048-2106',
                    '2137-2584',
                    '2615-2772',
                    '2823-2829',
                    '2880-3153',
                    '3354-4607',
                    '4866-5376',
                    '5632-6655',
                    '6912-7466',
                    '7723-8191',
                    '10240-12287',
                    '13312-15359',
                    '16384-17407',
                    '18432-20479',
                    '21504-23455',
                    '23457-23551',
                    '25600-26623',
                    '26624-27647',
                    '29696-30719',
                    '31744-32767',
                    '32768-33791',
                    '35840-36863',
                    '39936-40959',
                    '46080-47103',
                    '53248-54271',
                    '54272-55295',
                    '62464-63487',
                    '64198-64296',
                    '393216-394239',
                    '394240-395164',
                    '395165-396188',
                    '396189-397212',
                    '397213-398236',
                    '398237-399260',
                    '399261-400284',
                    '400285-401308'
                ],
                [
                    'https://rdap.arin.net/registry/',
                    'http://rdap.arin.net/registry/'
                ]
            ],
            [
                [
                    '1877-1901',
                    '2043',
                    '2047',
                    '2107-2136',
                    '2585-2614',
                    '2773-2822',
                    '2830-2879',
                    '3154-3353',
                    '5377-5631',
                    '6656-6911',
                    '8192-9215',
                    '12288-13311',
                    '15360-16383',
                    '20480-21503',
                    '24576-25599',
                    '28672-29695',
                    '30720-31743',
                    '33792-34815',
                    '34816-35839',
                    '38912-39935',
                    '40960-41983',
                    '41984-43007',
                    '43008-44031',
                    '44032-45055',
                    '47104-48127',
                    '48128-49151',
                    '49152-50175',
                    '50176-51199',
                    '51200-52223',
                    '56320-57343',
                    '57344-58367',
                    '59392-60415',
                    '60416-61439',
                    '61952-62463',
                    '64396-64495',
                    '196608-197631',
                    '197632-198655',
                    '198656-199679',
                    '199680-200191',
                    '200192-201215',
                    '201216-202239',
                    '202240-203263',
                    '203264-204287',
                    '204288-205211',
                    '205212-206235',
                    '206236-207259',
                    '207260-208283',
                    '208284-209307',
                    '209308-210331',
                    '210332-211355',
                    '211356-212379',
                    '212380-213403'
                ],
                [
                    'https://rdap.db.ripe.net/'
                ]
            ],
            [
                [
                    '27648-28671',
                    '52224-53247',
                    '61440-61951',
                    '64099-64197',
                    '262144-263167',
                    '263168-263679',
                    '263680-264604',
                    '264605-265628',
                    '265629-266652',
                    '266653-267676',
                    '267677-268700',
                    '268701-269724',
                    '269725-270748',
                    '270749-271772',
                    '271773-272796',
                    '272797-273820'
                ],
                [
                    'https://rdap.lacnic.net/rdap/'
                ]
            ]
        ],
        'version': '1.0'
    }
    __rirs = {
        'ripe': 'https://rdap.db.ripe.net',
        'arin': 'https://rdap.arin.net/registry',
        'afrinic': 'https://rdap.afrinic.net/rdap',
        'lacnic': 'https://rdap.lacnic.net/rdap',
        'apnic': 'https://rdap.apnic.net',
    }

    def __find_rir_by_asn(self, asn: int) -> str:
        for rir in self.__rdap_boot_strap['services']:
            url = rir[1][0]  # https rdap url
            for asns in rir[0]:
                asns = asns.split('-')
                if len(asns) == 1:
                    asns = asns + asns
                if int(asns[0]) <= asn <= int(asns[1]):
                    return url
        return self.__rir

    def __init__(self, default_rir: str, timeout: int):
        self.__timeout = timeout
        self.__rir = self.__rirs.get(default_rir, 'https://rdap.db.ripe.net')
        self.__known_asns = {}

    def get_whois_data_by_asn(self, asn: int) -> Dict[str, str]:
        asn_info = {}
        # don't fetch for private ASNs, two byte or four byte ASN number
        if asn < 64512 or (65536 < asn < 4200000000):
            rir = self.__find_rir_by_asn(asn)
            query = 'autnum'
            data = _fetch_data(f'{rir}{query}/{asn}', self.__timeout)
            if data:
                asn_info['as_name'] = data.get('name')
                vcard_array = data.get('entities')[0].get('vcardArray')
                if vcard_array:
                    for line in vcard_array[1]:
                        if line[0] == 'fn':
                            asn_info['as_org_name'] = line[3]
        self.__known_asns.update(asn_info)
        return asn_info
